<?php

class Truevo_API_Wrapper
{
	/**
     * How many times to attempt to fetch a transaction status before aborting.
     */
    const MAX_QUERY_STATUS_ATTEMPTS = 5;

    /**
     * How many attempts to get the transaction status have already been made.
     *
     * @var integer
     */
    private $attempts = 0;

    protected $idempotencyKey = null;
    protected $authorization = null;
    protected $apiUrl = null;
    protected $merchantId = null;
    protected $terminalId = null;

    public function __construct( $api_url, $authorization, $merchant_id, $terminal_id )
    {
        
        $this->apiUrl = $api_url;
        $this->authorization = $authorization;
        $this->idempotencyKey = $this->gen_uuid();
        $this->merchantId = $merchant_id;
        $this->terminalId = $terminal_id;
        
    }
    
    /**
     * gen_uuid function.
     * 
     * @access public
     * @return void
     */
    public function gen_uuid() {
	    return sprintf( '%04x%04x-%04x-%04x-%04x-%04x%04x%04x',
	        // 32 bits for "time_low"
	        mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff ),
	
	        // 16 bits for "time_mid"
	        mt_rand( 0, 0xffff ),
	
	        // 16 bits for "time_hi_and_version",
	        // four most significant bits holds version number 4
	        mt_rand( 0, 0x0fff ) | 0x4000,
	
	        // 16 bits, 8 bits for "clk_seq_hi_res",
	        // 8 bits for "clk_seq_low",
	        // two most significant bits holds zero and one for variant DCE1.1
	        mt_rand( 0, 0x3fff ) | 0x8000,
	
	        // 48 bits for "node"
	        mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff ), mt_rand( 0, 0xffff )
	    );
	}
    
    
    /**
     * get_query_status function.
     * 
     * @access public
     * @param mixed $data
     * @return void
     */
    public function get_query_status( $transaction_id )
    {
	    $params = [
			'transactionUniqueIdentifier' => $transaction_id,
			'merchantIdentifier' => $this->merchantId,
			'terminalIdentifier' => $this->terminalId,
		];

		return $this->_send_request('GET', '/payments/status', $params);
    }
    
    /**
     * get_payment_details function.
     * 
     * @access public
     * @param mixed $data
     * @return void
     */
    public function get_payment_details( $transaction_id, $is_available_capture = false, $log_enable = 'no' )
    {
		$log = new WC_Logger();
        $params = [
            'transactionUniqueIdentifier' => $transaction_id
        ];
		if ( $log_enable == 'yes' ) {
			$log->add('truevo_gateway', 'Truevo payments detail request prepared. ' . print_r($params, true));
		}
        $refund_mode = '';
        $get_payment_details = $this->_send_request('GET', '/payments/details', $params);
		
		if ( $log_enable == 'yes' ) {
			$log->add('truevo_gateway', 'Truevo payments details response. ' . print_r($get_payment_details, true));
		}
		
        if (
            $get_payment_details['code'] == '200'
            && !empty($get_payment_details['result'])
            && !empty($get_payment_details['result']->paymentDetails)
        ) {
            // Build missing properties
            if (!property_exists($get_payment_details['result']->paymentDetails, 'isCaptureAllowed'))    $get_payment_details['result']->paymentDetails->isCaptureAllowed = false;
            if (!property_exists($get_payment_details['result']->paymentDetails, 'isRefundAllowed'))     $get_payment_details['result']->paymentDetails->isRefundAllowed = false;
            if (!property_exists($get_payment_details['result']->paymentDetails, 'isReverseAllowed'))    $get_payment_details['result']->paymentDetails->isReverseAllowed = false;
            if (!property_exists($get_payment_details['result']->paymentDetails, 'availableForCapture')) $get_payment_details['result']->paymentDetails->availableForCapture = 0;
            if (!property_exists($get_payment_details['result']->paymentDetails, 'availableForRefund'))  $get_payment_details['result']->paymentDetails->availableForRefund = 0;

            if ($is_available_capture) {
                if ($get_payment_details['result']->paymentDetails->isCaptureAllowed) {
                    $refund_mode = 'capture';
                }
            } else {
                if ($get_payment_details['result']->paymentDetails->isRefundAllowed && $get_payment_details['result']->paymentDetails->availableForRefund > 0) {
                    $refund_mode = 'refund';
                } elseif ($get_payment_details['result']->paymentDetails->isReverseAllowed) {
                    $refund_mode = 'reverse';
                }
            }
        }
		
		if ( $log_enable == 'yes' ) {
			$log->add('truevo_gateway', 'Truevo payments details mode returned. ' . print_r($refund_mode, true));
		}
		
        return $refund_mode;
    }
	    
	    /**
     * process_truevo_refund function.
     * 
     * @access public
     * @param mixed $data
     * @return void
     */
    public function process_truevo_refund( $data, $mode )
    {
	    
		$params = [
			'transaction' => [
				'originalTransactionUniqueIdentifier' => $data['transactionUniqueIdentifier'],
				'merchantReference' => $data['merchantReference'],
				'paymentReference' => $data['paymentReference'],
				'amount' => [
					'totalAmount' => $data['amount'],
					'currencyAlphaCode' => get_woocommerce_currency(),
					'details' => [
						'paymentAmount' => $data['amount'],
						'gratuityAmount' => 0,
					],
				],
				'source' => 'ECOMMERCE'
			],
			'merchant' => [
				'mid' => $this->merchantId,
				'tid' => $this->terminalId
			],
			'customer' => [
				'firstName' => $data['billing_first_name'],
				'lastName' => $data['billing_last_name'],
				'contact' => [
					'email' => $data['email'],
					'ip' => $data['user_ip'],
				],
				'physicalAddress' => [
					'street1' => $data['delivery_address'],
					'street2' => $data['delivery_address_2'],
					'city' => $data['delivery_city'],
					'state' => $data['delivery_state'],
					'postCodeZip' => $data['delivery_postcode'],
					'countryAlphaCode' => $data['delivery_country'],
				],
				'billingAddress' => [
					'street1' => $data['billing_address'],
					'street2' => $data['billing_address_2'],
					'city' => $data['billing_city'],
					'state' => $data['billing_state'],
					'postCodeZip' => $data['billing_postcode'],
					'countryAlphaCode' => $data['billing_country'],
				]
			],
			'card' => [],
			'additionalData' => [],
		];

        if ($mode == 'refund') {
            return $this->_send_request('POST', '/payments/linkedRefund', $params);
        } else {
            return $this->_send_request('POST', '/payments/reversal', $params);
        }
		
    }
    
    public function process_truevo_capture( $data )
    {
	    
		$params = [
			'transaction' => [
				'originalTransactionUniqueIdentifier' => $data['transactionUniqueIdentifier'],
				'retrievalReferenceNumber' => $data['retrievalReferenceNumber'],
				'merchantReference' => $data['merchantReference'],
				'paymentReference' => $data['paymentReference'],
				'amount' => [
					'totalAmount' => $data['amount'],
					'currencyAlphaCode' => get_woocommerce_currency(),
					'details' => [
						'paymentAmount' => $data['amount'],
						'gratuityAmount' => 0,
					],
				],
				'source' => 'ECOMMERCE'
			],
			'merchant' => [
				'mid' => $this->merchantId,
				'tid' => $this->terminalId
			],
			'customer' => [
				'firstName' => $data['billing_first_name'],
				'lastName' => $data['billing_last_name'],
				'contact' => [
					'email' => $data['email'],
					'ip' => $data['user_ip'],
				],
				'physicalAddress' => [
					'street1' => $data['delivery_address'],
					'street2' => $data['delivery_address_2'],
					'city' => $data['delivery_city'],
					'state' => $data['delivery_state'],
					'postCodeZip' => $data['delivery_postcode'],
					'countryAlphaCode' => $data['delivery_country'],
				],
				'billingAddress' => [
					'street1' => $data['billing_address'],
					'street2' => $data['billing_address_2'],
					'city' => $data['billing_city'],
					'state' => $data['billing_state'],
					'postCodeZip' => $data['billing_postcode'],
					'countryAlphaCode' => $data['billing_country'],
				]
			],
			'additionalData' => [],
		];
		
		return $this->_send_request('POST', '/payments/capture', $params);
    }
    

    
    /**
     * Prepares and sends your request to the API servers
     *
     * @param string $method POST | GET | UPDATE | DELETE
     * @param string $action Truevo API endpoint
     * @param array $data A Truevo API Request Body packet as an array
     * @return boolean|array Returns Exception/FALSE or the "Response" array
   */
    public function _send_request($method, $action, $data = null)
    {
		
        $endpoint = $this->apiUrl;
        $idempotencyKey = $this->idempotencyKey;
        $superSecretCode = $this->authorization;
        
        $timestamp = gmdate("c");
        $auth = "Authorization: Bearer $superSecretCode";
        
        $auth_key = '';
        if ($method == 'POST') {
			$auth_key = "idempotencyKey : $idempotencyKey";
		}
		
		$url = $endpoint.$action;
		
		if ($method == 'GET') {
			$url = add_query_arg( 
			    $data, 
			    $url
			);
		}
		
        
        $post_args = json_encode($data);
        $headers = array($auth, $auth_key, "Accept: application/json", "Content-Type: application/json; charset=utf-8");
		
        $curl = curl_init();
        curl_setopt($curl, CURLOPT_URL, $url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_ENCODING, "");
        curl_setopt($curl, CURLOPT_MAXREDIRS, 10);
        curl_setopt($curl, CURLOPT_TIMEOUT, 30);
        curl_setopt($curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
        
        
        curl_setopt($curl, CURLOPT_CUSTOMREQUEST, $method);
        
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);
        
        if( $method == "POST" ){
	        curl_setopt($curl, CURLOPT_POST, true);
			curl_setopt($curl, CURLOPT_POSTFIELDS, $post_args);
        }
        
        
        $result = curl_exec($curl);
        $responseCode = curl_getinfo($curl, CURLINFO_HTTP_CODE);
                
        $result = ['code' => $responseCode, 'result' => json_decode($result)];
		
		if ($method == 'PUT') {
            return $responseCode;
        }
		
        curl_close($curl);
        
        
        return $result;
    }

	/**
     * Poll the status for a transaction unique identifier until either a 
     * "COMPLETE" or "DECLINED" is returned or the request times out.
     *
     * @param string $transaction_id
     * @param string $log_enable
     * @return array Contains 'status' and 'reason', 'status' should be one of the following: "COMPLETE", "DECLINED", or "PENDING"
     */
	public function poll_transaction_status( $transaction_id, $log_enable = 'no' )
	{
        $log = new WC_Logger();

        $is_complete = false;
        $return = [
            'status' => 'PENDING',
            'reason' => 'Request timed out'
        ];

        do {
            sleep(1); // Wait a second before querying the status
            $this->attempts += 1; // Increment the total number of attempts

            $response = $this->get_query_status($transaction_id);

            if ($response['result']) {
                $return['status'] = $response['result']->outcome->transactionStatus;

                // Don't log pending requests unless it is the final one
                if ($return['status'] !== "PENDING" || $this->attempts === self::MAX_QUERY_STATUS_ATTEMPTS) {
                    $return['reason'] = $response['result']->outcome->reasonDescription;

                    if ( $log_enable == 'yes' ) {
                        $log->add('truevo_gateway', 'Truevo poll transaction status response. ' . print_r($response['result'], true));
                    }
                }
                
                $is_complete = $return['status'] !== "PENDING"; 
            }
        } while ($this->attempts < self::MAX_QUERY_STATUS_ATTEMPTS && !$is_complete);

        return $return;
	}
}
